//************************************************************
//  University of Patras 
//	Department of Electrical Engineering & Computer Technology
//	 
//	Course: Parallel Programming in AI 
//	Week 03: Collatz conjecture
//
//  Description: How many steps is required starting from an arbitrary
//                  natural number to reach 1?
//        The collatz algorithm: 
//          Start from the number n
//          if n is even then n <- n/2 else n <- 3n+1
//          if n equal 1 stop else repeat previous if            
//
//      example: 12, 6, 3, 10, 5, 16, 8, 4, 2, 1.
//
//  	  collatz01.c: estimate the most frequent number of steps reaching 1 using the collatz rule  
//	Author: Evangelos Dermatas
//	v0.1
/*
*********************************************************************
gcc -v:	gcc version 7.4.0 (Ubuntu 7.4.0-1ubuntu1~18.04.1)
OS: 	Linux Ubuntu 18.04
CPU:	8-core, Intel(R) Core(TM) i7-2630QM CPU @ 2.00GHz, 3991 BogoMips
*********************************************************************
Compiling:	gcc collatz01.c -o collatz01
Executing:	time ./collatz01

Output in my Laptop:

Too long to wait ....

*********************************************************************
Compiling:	gcc collatz01.c -o collatz01 -O2
Executing:	time ./collatz01

Output in my Laptop:

Steps 174, Frequency 994530

real	6m2.681s
user	6m2.669s
sys	    0m0.004s

*/

// Collatz assumption ...
#include <stdio.h>
#include <stdlib.h>

// Maximum number of steps
#define NSTEPS  100000
// Maximum natural number

#define N       100000000
unsigned long int Hist[NSTEPS+1] ;


// ***************************************
void CollatzSerial(void)
{
    int q, k ;
    for( int i = 1 ; i < N ; i++  ) {
        for ( k = 0, q = i ; k<NSTEPS ; k++ ) {
            if ( q == 1 ) {
                Hist[k]++ ;
                break ;
            }
            if ( q%2 == 0 )
                q /= 2 ;
            else
                q=3*q+1;
        }
        if ( k == NSTEPS )
            Hist[NSTEPS]++ ;
    }
}

// ***************************************
void PrintResults() 
{
    unsigned long int max=0 ;
    int ind=0;

for ( int k = 0 ; k < NSTEPS ; k++ )
    if ( max < Hist[k] ) {
        ind = k ;
        max = Hist[k] ;
    }
    printf( "Steps %d, Frequency %lu\n", ind, max) ;
}


// ***************************************
int main()
{
    CollatzSerial() ;
    PrintResults() ;
    return 0 ;
}
